use std::io::Write;
use std::net::TcpStream;
use std::os::fd::AsRawFd;
use base64::Engine;
use hhmmss::Hhmmss;
use crate::{Data, service};

fn index_write_service(stream: &mut impl Write, service: &crate::service::Service) -> Option<()> {
    let (state, desc, can_start, can_stop) = match &service.status {
        crate::service::ServiceStatus::STOPPED => ("stopped", "".to_string(), true, false),
        crate::service::ServiceStatus::STARTED(started_at) =>
            ("started", format!("Pid: {}, uptime: {}", service.pid.as_ref().map_or(-1, |v| v.0.as_raw()), started_at.elapsed().hhmmss()), false, true),
        crate::service::ServiceStatus::STOPPING => ("stopping", "".to_string(), false, false),
        crate::service::ServiceStatus::FAILED(err) => ("failed", err.clone(), true, false)
    };
    write!(stream, "<tr><td data-status=\"{}\"></td><td>{}</td><td>{}</td><td>", state, &service.config.name, desc).ok()?;
    if can_start { write!(stream, "<a href=\"/start/{}\">Start</a>", &service.config.name).ok()?; }
    stream.write_all(b"</td><td>").ok()?;
    if can_stop { write!(stream, "<a href=\"/stop/{}\">Stop</a>", &service.config.name).ok()?; }
    write!(stream, "</td><td><a href=\"/tail#{}\">Tail log</a></td></tr>", &service.config.name).ok()
}

pub fn send_404(mut stream: TcpStream) -> Option<()> {
    stream.write_all(b"HTTP/1.0 404 Not found\r\nContent-Type: text/plain;charset=utf-8\r\n\r\n404 Not found").ok()
}

pub fn send_301(mut stream: TcpStream) -> Option<()> {
    stream.write_all(b"HTTP/1.0 302 Found\r\nLocation: /\r\n\r\n").ok()
}

pub fn send_index(stream: TcpStream) -> Option<()> {
    let mut stream = std::io::BufWriter::new(stream);
    stream.write_all(b"HTTP/1.0 200 OK\r\nContent-Type: text/html;charset=utf-8\r\n\r\n").ok()?;
    stream.write_all(b"<!DOCTYPE html><html lang=\"en\"><head><title>Minitd</title><style>
        .center-child { display: flex; flex-direction: column; align-items: center; row-gap: 2em; background-color: #aaffaa44; }
        .space-child > * { margin: 0 0.75em; }
        a, a:link, a:visited { color: #444; }
        a:hover { color: #000; }
        table { border: 1px solid #999; }
        thead { color: #fff; background-color: #999; }
        td, th { padding: 0.25em 0.5em; border-bottom: 1px solid #fff; }
        tbody > tr:nth-child(even) { background-color: #eee; }
        tbody > tr:nth-child(odd) { background-color: #ddd; }
        [data-status] { padding: 0.125em 0.25em; }
        [data-status]::after { padding: 0.125em 0.25em; }
        [data-status='stopped']::after { content: 'Stopped'; background-color: #ef8; }
        [data-status='started']::after { content: 'Started'; background-color: #8f8; }
        [data-status='stopping']::after { content: 'Stopping'; background-color: #fc8; }
        [data-status='failed']::after { content: 'Failed'; background-color: #faa; }
    </style></head><body class=\"center-child\">").ok()?;
    stream.write_all(b"<div class=\"space-child\"><a href=\".\">Refresh</a><a href=\"/tail#minitd\">Tail log</a><a href=\"/reload\">Reload config</a><a href=\"/stop_all\">Stop all</a><a href=\"/shutdown\">Shutdown</a></div>").ok()?;
    stream.write_all(b"<table cellspacing=\"0\"><thead><th>State</th><th>Name</th><th>Description</th><th colspan=\"3\">Action</th></thead><tbody>").ok()?;
    for (_, service) in &crate::Data::get().services {
        index_write_service(&mut stream, service)?;
    }
    stream.write_all(b"</tbody></table></body></html>").ok()
}

pub fn send_tail_html(stream: TcpStream) -> Option<()> {
    let mut stream = std::io::BufWriter::new(stream);
    stream.write_all(b"HTTP/1.0 200 OK\r\nContent-Type: text/html;charset=utf-8\r\n\r\n").ok()?;
    stream.write_all(b"<!DOCTYPE html><html lang=\"en\"><head><title>Minitd</title></head><body><pre id=\"c\"></pre><script>").ok()?;
    stream.write_all(b"const c = document.getElementById('c'); const e = new EventSource('/_tail/' + location.hash.slice(1)); e.onerror = _ => e.close(); e.addEventListener('line', msg => c.innerText += atob(msg.data));").ok()?;
    stream.write_all(b"</script></body></html>").ok()
}

pub fn send_tail(mut stream: TcpStream, file: String) -> Option<()> {
    let mut b64_buf = [0_u8; 2048];
    stream.write_all(b"HTTP/1.0 200 OK\r\ncontent-type: text/event-stream\r\n\r\n").ok()?;

    let file = match std::fs::File::open(file) {
        Ok(v) => v,
        Err(err) => {
            let msg = format!("Failed to open file: {err}");
            let size = base64::prelude::BASE64_STANDARD.encode_slice(&msg, &mut b64_buf).ok()?;
            stream.write_all(b"event:line\ndata:").ok()?;
            stream.write_all(&b64_buf[..size]).ok()?;
            return stream.write_all(b"\n\n").ok();
        }
    };
    let mut buf = [0_u8; 1024];
    let mut fds = [nix::poll::PollFd::new(&file, nix::poll::PollFlags::POLLIN | nix::poll::PollFlags::POLLHUP | nix::poll::PollFlags::POLLERR)];

    loop {
        let ready = nix::poll::poll(&mut fds, 15000).ok()?;
        if ready == 0 { // timeout
            stream.write_all(b"event:ka\ndata:\n\n").ok()?;
        } else if fds[0].revents().unwrap().contains(nix::poll::PollFlags::POLLIN) {
            let size = nix::unistd::read(file.as_raw_fd(), &mut buf).ok()?;
            if size > 0 {
                let size = base64::prelude::BASE64_STANDARD.encode_slice(&buf[..size], &mut b64_buf).ok()?;
                stream.write_all(b"event:line\ndata:").ok()?;
                stream.write_all(&b64_buf[..size]).ok()?;
                stream.write_all(b"\n\n").ok()?;
            }
        } else {
            break
        }
    }

    None
}

pub fn reload_config() -> Result<(), String> {
    let data = Data::get();
    let new_config = crate::config::Config::load()?;
    let mut removed_services = data.services.keys().cloned().collect::<std::collections::HashSet<_>>();
    for service in new_config.service {
        match data.services.iter_mut().find(|s| s.1.config.name == service.name) {
            Some((id, s)) => { // Replace old service
                s.config = service;
                removed_services.remove(id);
            }
            None => { // New service
                let s = service::Service::new(service);
                data.services.insert(s.id, s);
            }
        }
    }

    let queue = data.work_queue.as_ref().unwrap();
    for work in removed_services.iter().map(|id| crate::WorkItem::StopService(*id, false)) { queue.send(work).unwrap(); }
    for id in removed_services {
        let s = data.services.get(&id).unwrap();
        while !s.stopped() { std::thread::sleep(std::time::Duration::from_millis(250)); }
        data.services.remove(&id);
    }

    Ok(())
}
