use std::os::unix::prelude::CommandExt;
use std::process::Command;
use crate::web::setup_webserver;

mod config;
mod service;
mod watch;
mod web;
mod html;

#[derive(Debug)]
enum WorkItem {
    StartService(u64),
    StopService(u64, bool)
}

#[derive(Debug)]
struct Data {
    pub epoll: nix::sys::epoll::Epoll,
    pub sigfd: nix::sys::signalfd::SignalFd,
    pub work_queue: Option<std::sync::mpsc::Sender<WorkItem>>,
    pub services: linked_hash_map::LinkedHashMap<u64, service::Service>
}
impl Data {
    fn new() -> Self {
        let epoll = nix::sys::epoll::Epoll::new(nix::sys::epoll::EpollCreateFlags::EPOLL_CLOEXEC).expect("Failed to create epoll");
        let mut sigset = nix::sys::signal::SigSet::empty();
        sigset.add(nix::sys::signal::SIGTERM);
        sigset.add(nix::sys::signal::SIGINT);
        sigset.add(nix::sys::signal::SIGCHLD);
        nix::sys::signal::sigprocmask(nix::sys::signal::SigmaskHow::SIG_BLOCK, Some(&sigset), None).expect("Failed to set signal mask");
        let sigfd = nix::sys::signalfd::SignalFd::with_flags(&sigset, nix::sys::signalfd::SfdFlags::SFD_CLOEXEC).expect("Failed to create signalfd");
        epoll.add(&sigfd, nix::sys::epoll::EpollEvent::new(nix::sys::epoll::EpollFlags::EPOLLIN, watch::EPOLL_SIGFD_ID)).expect("Failed to add signalfd to epoll");
        Self {
            epoll,
            sigfd,
            work_queue: None,
            services: Default::default()
        }
    }

    pub fn get() -> &'static mut Self {
        static mut INSTANCE: Option<Data> = None;
        unsafe {
            if INSTANCE.is_none() { INSTANCE.replace(Data::new()); }
            INSTANCE.as_mut().unwrap()
        }
    }

    pub fn any_running(&self) -> bool {
        return self.services.iter().any(|s| !s.1.stopped());
    }

    pub fn service_update_dep(&mut self) {
        let mut id_map = std::collections::HashMap::new();
        for (_, service) in &self.services {
            id_map.insert(service.config.name.clone(), service.id);
        }

        let id_map = id_map;
        let mut dep_map: std::collections::HashMap<u64, Vec<u64>> = std::collections::HashMap::new();
        for (_, service) in &mut self.services {
            service.dependencies.clear();
            for dep in &service.config.depends_on {
                let did = *id_map.get(dep).unwrap();
                service.dependencies.push(did);
                dep_map.entry(did).or_default().push(service.id);
            }
        }
        for (_, service) in &mut self.services {
            service.dependents = dep_map.entry(service.id).or_default().clone();
        }
    }
}

fn main() {
    let (start_as_daemon, config) = {
        let mut start_as_daemon = true;

        let mut arg_iter = std::env::args().skip(1);
        while let Some(arg) = arg_iter.next() {
            if arg == "-n" {
                start_as_daemon = false;
            } else if arg == "-s" {
                let new_args = std::env::args().filter(|s| s != "-s");
                Command::new("strace").arg("-f").arg("-e").arg("trace=%process,%ipc,/epoll.*").args(new_args).exec();
            } else if arg.starts_with('-') {
                eprintln!("Invalid switch '{arg}'");
                std::process::exit(1);
            } else {
                eprintln!("Invalid arg '{arg}'");
                std::process::exit(1);
            }
        }

        let config = match config::Config::load() {
            Ok(v) => v,
            Err(e) => { eprintln!("{e}"); std::process::exit(1); }
        };

        (start_as_daemon, config)
    };

    if start_as_daemon {
        println!("Launching as daemon!");
        daemonize::Daemonize::new()
            .working_directory(std::env::current_dir().unwrap())
            .start()
            .expect("Failed to daemonize");
        let log_file = nix::fcntl::open(
            "minitd.log",
            nix::fcntl::OFlag::O_WRONLY | nix::fcntl::OFlag::O_CLOEXEC | nix::fcntl::OFlag::O_CREAT | nix::fcntl::OFlag::O_APPEND,
            nix::sys::stat::Mode::S_IWUSR | nix::sys::stat::Mode::S_IRUSR
        ).expect("Failed to open log file");
        nix::unistd::dup2(log_file, 1).expect("Failed to redirect stdout");
        nix::unistd::dup2(log_file, 2).expect("Failed to redirect stderr");
        nix::unistd::close(log_file).expect("Failed to close log file");
    }

    {
        let data = Data::get();
        for service in config.service {
            let service = service::Service::new(service);
            data.services.insert(service.id, service);
        }
        data.service_update_dep();
    }

    setup_webserver();

    watch::watch_services(config.autostart);
}
