use std::ffi::CString;
use std::os::fd::{AsRawFd, FromRawFd};
use crate::Data;

#[derive(Debug, PartialEq)]
pub enum ServiceStatus {
    STOPPED,
    STARTED(std::time::Instant), // Started at
    STOPPING,
    FAILED(String)
}

#[derive(Debug)]
pub struct Service {
    pub id: u64, // > 255
    pub pid: Option<(nix::unistd::Pid, std::os::fd::OwnedFd)>,
    pub status: ServiceStatus,
    pub reaped: std::sync::atomic::AtomicBool,
    pub dependencies: Vec<u64>,
    pub dependents: Vec<u64>,
    pub restart_count: u8,
    pub config: crate::config::CService
}

impl Service {
    fn get_args(&self) -> Result<Vec<CString>, String> {
        let args = match shlex::split(&self.config.command) {
            Some(v) => v,
            None => return Err("Invalid command".into())
        };
        if args.is_empty() {
            return Err("Command is empty".into());
        }
        Ok(args.into_iter().map(|s| CString::new(s).unwrap()).collect::<Vec<_>>())
    }

    pub fn started(&self) -> bool {
        match &self.status {
            ServiceStatus::STARTED(_) => !self.reaped.load(std::sync::atomic::Ordering::Relaxed),
            _ => false
        }
    }

    pub fn stopped(&self) -> bool {
        match &self.status {
            ServiceStatus::STOPPED | ServiceStatus::FAILED(_) => true,
            _ => false
        }
    }
    
    pub fn can_autostart(&self) -> bool {
        return self.restart_count < 3;
    }

    pub fn new(config: crate::config::CService) -> Self {
        static NEXT_SERVICE_ID: std::sync::atomic::AtomicU64 = std::sync::atomic::AtomicU64::new(256);

        let id = NEXT_SERVICE_ID.fetch_add(1, std::sync::atomic::Ordering::Relaxed);
        Self {
            id,
            pid: None,
            status: ServiceStatus::STOPPED,
            reaped: std::sync::atomic::AtomicBool::new(false),
            dependents: Vec::new(),
            dependencies: Vec::new(),
            restart_count: 0,
            config
        }
    }

    pub fn start(&mut self) {
        if self.started() { return; }
        println!("[U...] {}", &self.config.name);
        self.restart_count += 1;
        self.reaped.store(false, std::sync::atomic::Ordering::Relaxed);

        let args = match self.get_args() {
            Ok(v) => v,
            Err(err) => {
                self.status = ServiceStatus::FAILED(err);
                println!("[FAIL] {}", &self.config.name);
                return;
            }
        };

        let (r, s) = nix::unistd::pipe().unwrap();
        let (r, s) = unsafe { (std::os::fd::OwnedFd::from_raw_fd(r), std::os::fd::OwnedFd::from_raw_fd(s)) };
        let mut pipe_buf = [0_u8];

        match unsafe {nix::unistd::fork()} {
            Ok(nix::unistd::ForkResult::Child) => {
                drop(r);
                let mut sigset = nix::sys::signal::SigSet::empty();
                sigset.add(nix::sys::signal::SIGTERM);
                sigset.add(nix::sys::signal::SIGINT);
                sigset.add(nix::sys::signal::SIGCHLD);
                nix::sys::signal::sigprocmask(nix::sys::signal::SigmaskHow::SIG_UNBLOCK, Some(&sigset), None).expect("Failed to set signal mask");

                let log_name = self.config.name.clone() + ".log";
                let log_file = nix::fcntl::open(
                    log_name.as_str(),
                    nix::fcntl::OFlag::O_WRONLY | nix::fcntl::OFlag::O_CLOEXEC | nix::fcntl::OFlag::O_CREAT | nix::fcntl::OFlag::O_APPEND,
                    nix::sys::stat::Mode::S_IWUSR | nix::sys::stat::Mode::S_IRUSR
                ).expect("Failed to open log file");
                nix::unistd::dup2(log_file, 1).expect("Failed to redirect stdout");
                nix::unistd::dup2(log_file, 2).expect("Failed to redirect stderr");
                nix::unistd::close(log_file).expect("Failed to close log file");
                nix::unistd::close(0).expect("Failed to close stdin");

                nix::unistd::setpgid(nix::unistd::Pid::from_raw(0), nix::unistd::Pid::from_raw(0)).expect("Failed to set pgid");

                if let Some(dir) = &self.config.directory { nix::unistd::chdir(dir.as_str()).expect(&format!("Failed to change dir to {dir}")); }

                nix::unistd::write(s.as_raw_fd(), &pipe_buf).unwrap();
                drop(s);

                nix::unistd::execvp(&args[0], args.as_slice()).unwrap();
                std::process::exit(0);
            }
            Ok(nix::unistd::ForkResult::Parent { child }) => {
                drop(s);

                let pidfd = nix::errno::Errno::result(unsafe { libc::syscall(libc::SYS_pidfd_open, child.as_raw(), 0 as libc::c_int) as libc::c_int }).expect("Failed to open pidfd");
                let pidfd = unsafe { std::os::fd::OwnedFd::from_raw_fd(pidfd) };
                self.pid.replace((child, pidfd));

                let mut fds = [
                    nix::poll::PollFd::new(&r, nix::poll::PollFlags::POLLIN | nix::poll::PollFlags::POLLHUP | nix::poll::PollFlags::POLLPRI),
                    nix::poll::PollFd::new(&r, nix::poll::PollFlags::POLLIN | nix::poll::PollFlags::POLLPRI)
                ];
                nix::poll::poll(&mut fds, -1).expect("Failed to poll fds");
                if fds[0].any().unwrap_or(false) {
                    let read = nix::unistd::read(r.as_raw_fd(), &mut pipe_buf).unwrap();
                    if read == 0 {
                        self.status = ServiceStatus::FAILED("Check log".into());
                        println!("[FAIL] {}", &self.config.name);
                        return;
                    }
                } else {
                    self.status = ServiceStatus::FAILED("Check log".into());
                    println!("[FAIL] {}", &self.config.name);
                    return;
                }

                self.status = ServiceStatus::STARTED(std::time::Instant::now());
                Data::get().epoll.add(&self.pid.as_ref().unwrap().1, nix::sys::epoll::EpollEvent::new(nix::sys::epoll::EpollFlags::EPOLLIN, self.id)).expect("Failed to add epoll");
                println!("[ OK ] {}", &self.config.name);
            }
            Err(err) => panic!("Failed to fork for service: {}", err)
        }
    }

    pub fn stop(&mut self, manual: bool) {
        const TERM_WAIT_MS: u128 = 10000;

        if self.stopped() { return; }

        if let Some((pid, _)) = self.pid.as_ref() {
            let pid = *pid;
            println!("[D...] {}", &self.config.name);
            self.status = ServiceStatus::STOPPING;
            let stop_started = std::time::Instant::now();
            nix::sys::signal::kill(pid, nix::sys::signal::SIGTERM).expect("Failed to send SIGTERM");
            while !self.reaped.load(std::sync::atomic::Ordering::Relaxed) && stop_started.elapsed().as_millis() < TERM_WAIT_MS {
                std::thread::sleep(std::time::Duration::from_millis(250));
            }
            if !self.reaped.load(std::sync::atomic::Ordering::Relaxed) {
                nix::sys::signal::kill(pid, nix::sys::signal::SIGKILL).expect("Failed to send SIGKILL");
            }
        }

        println!("[STOP] {}", &self.config.name);
        self.status = ServiceStatus::STOPPED;

        if manual { self.restart_count = 0; }
    }
}
