use std::io::{Write, BufRead};
use std::net::TcpStream;
use crate::{Data, html, watch, WorkItem};

static mut WEBSERVER_FD: Option<std::net::TcpListener> = None;

pub fn setup_webserver() {
    let data = Data::get();
    let listener = std::net::TcpListener::bind("0.0.0.0:9001").expect("Failed to bind listener");
    listener.set_nonblocking(true).expect("Failed to set non-blocking");
    data.epoll.add(&listener, nix::sys::epoll::EpollEvent::new(nix::sys::epoll::EpollFlags::EPOLLIN, watch::EPOLL_WEBSERVER_ID)).expect("Failed to add socket to epoll");
    unsafe { WEBSERVER_FD = Some(listener) };
}

pub fn webserver_accept() {
    let listener = unsafe { WEBSERVER_FD.as_ref().unwrap() };
    if let Ok((stream, _)) = listener.accept() {
        std::thread::spawn(move || handle_stream(stream));
    }
}

fn handle_stream(mut stream: TcpStream) -> Option<()> {
    stream.set_read_timeout(Some(std::time::Duration::new(5, 0))).ok()?;
    stream.set_write_timeout(Some(std::time::Duration::new(5, 0))).ok()?;

    let reader = std::io::BufReader::new(&mut stream);
    let mut lines = reader.lines();
    let (method, path) = {
        let first = lines.next()?.ok()?;
        let (method, first) = first.split_once(' ')?;
        let path = first.split_once(' ')?.0;
        (method.to_string(), path.to_string())
    };
    for line in lines {
        if line.ok()?.is_empty() { break; }
    }
    if method != "GET" { return html::send_404(stream); }

    if path == "/" {
        html::send_index(stream)
    } else if path == "/tail" {
        html::send_tail_html(stream)
    } else if path == "/reload" {
        match html::reload_config() {
            Ok(()) => html::send_301(stream),
            Err(e) => write!(stream, "HTTP/1.0 400 Bad request\r\nContent-Type: text/plain;charset=utf-8\r\n\r\nFailed to reload config:\n{e}").ok()
        }
    }  else if path == "/stop_all" {
        let data = Data::get();
        for (id, _) in &data.services {
            data.work_queue.as_ref().unwrap().send(WorkItem::StopService(*id, false)).unwrap();
        }
        html::send_301(stream)
    } else if path == "/shutdown" {
        nix::sys::signal::kill(nix::unistd::getpid(), nix::sys::signal::SIGTERM).ok()?;
        html::send_301(stream)
    } else if path.starts_with("/_tail/") {
        let mut path = path;
        drop(path.drain(..7));
        if path.contains('.') || path.contains('/') {
            html::send_404(stream)
        } else {
            let path = path + ".log";
            html::send_tail(stream, path)
        }
    } else if path.starts_with("/start/") {
        let service = &path[7..];
        let data = Data::get();
        match data.services.iter().find(|(_, s)| s.config.name == service) {
            Some((id, _)) => {
                data.work_queue.as_ref().unwrap().send(WorkItem::StartService(*id)).unwrap();
                html::send_301(stream)
            },
            None => html::send_404(stream)
        }
    } else if path.starts_with("/stop/") {
        let service = &path[6..];
        let data = Data::get();
        match data.services.iter().find(|(_, s)| s.config.name == service) {
            Some((id, _)) => {
                data.work_queue.as_ref().unwrap().send(WorkItem::StopService(*id, true)).unwrap();
                html::send_301(stream)
            },
            None => html::send_404(stream)
        }
    } else {
        html::send_404(stream)
    }
}
